############### Copyright 2004-2016 Team GoldenEye: Source #################
#
# This file is part of GoldenEye: Source's Python Library.
#
# GoldenEye: Source's Python Library is free software: you can redistribute 
# it and/or modify it under the terms of the GNU General Public License as 
# published by the Free Software Foundation, either version 3 of the License, 
# or(at your option) any later version.
#
# GoldenEye: Source's Python Library is distributed in the hope that it will 
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General 
# Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GoldenEye: Source's Python Library.
# If not, see <http://www.gnu.org/licenses/>.
#############################################################################

# TABLE OF CONTENTS
# 1 - IMPORTS
# 2 - SYSTEM
# 3 - DATA
#   3.1 - Tracker Constants
#   3.2 - Command List
#   3.3 - Message Text
#   3.4 - Color Constants
#   3.5 - Message Channels
#   3.6 - Hud Bar Indexes
#   3.7 - Weapon Data
#
# 4 - FUCNTIONS
#   4.1 - Implication
#
#   4.2 - CLASS FUNCTIONS
#       4.2.1  - __init__
#       4.2.2  - GetPrintName
#       4.2.3  - GetScenarioHelp
#       4.2.4  - GetGameDescription
#       4.2.5  - GetTeamPlay
#       4.2.6  - OnLoadGameplay
#       4.2.7  - OnUnloadGamePlay
#       4.2.8  - OnCVarChanged
#       4.2.9  - OnRoundBegin
#       4.2.10 - OnPlayerSpawn
#       4.2.11 - OnPlayerKilled
#       4.2.12 - OnPlayerSay
#       4.2.13 - OnThink
#       4.2.14 - CanMatchEnd
#
#   4.3 - UTILITY FUNCTIONS
#       4.3.1 - u_GetRandom
#       4.3.2 - u_CheckRange
#       4.3.3 - u_CreateCVars
#       4.3.4 - u_CheckKillEnd
#       4.3.5 - u_GetDigit
#
#   4.4 - TIMER FUNCTIONS
#       4.4.1 - t_OffsetTime
#       4.4.2 - t_CheckTimer
#
#   4.5 - PLAYER FUNCTION
#       4.5.1 - p_ResetPlayer
#       4.5.2 - p_OnKilled
#       4.5.3 - p_OnKill
#
#   4.6 - WEPAON FUNCTIONS
#       4.6.1   - wp_GiveWeapons
#       4.6.2   - wp_AssignWeapons
#       4.6.3   - wp_AssignSingleWeapon
#       4.6.4   - wp_GenWeaponCat
#       4.6.5   - wp_GenWeapon
#       4.6.6   - wp_StealthWeapons
#       4.6.7   - wp_SpecialWeapons
#       4.6.8   - wp_SetCarryLoadout
#
#   4.7 - WEAPON SWITCHING FUNCTIONS
#       4.7.1  - sw_AwardBonus
#       4.7.2  - sw_GiveAwardSwitch
#       4.7.3  - sw_SwitchWeapons
#       4.7.4  - sw_PerformSwitch
#       4.7.5  - sw_RollingSwitch
#       4.7.6  - sw_GetSwitchDelay
#       4.7.7  - sw_GetRollingSwitchDelay
#       4.7.8  - sw_SlaperKill
#       4.7.9  - sw_SetCarry
#       4.7.10 - sw_CheckCarry
#       4.7.11 - sw_SwitchArmor
#
#   4.8 - UI FUNCTIONS
#       4.8.1 - ui_CreateSwitchBar
#       4.8.2 - ui_UpdateSwitchBar
#       4.8.3 - ui_DestroySwitchBar
#       4.8.4 - ui_GetSwitchBarContent
#       4.8.5 - ui_PrintLoadout
#       4.8.6  - ui_NotifyKillsRemaining

# <1> IMPORTS
import sys
import math
import random
import GEPlayer, GEUtil, GEMPGameRules as GERules, GEGlobal as Glb, GEWeapon

from GamePlay import GEScenario
from .Utils import GetPlayers, _
from .Utils.GEPlayerTracker import GEPlayerTracker
from GEUtil import GetTime
from string import Template as Template
from functools import partial

# <2> SYSTEM
USING_API = Glb.API_VERSION_1_2_0

# <3> DATA
# <3.1> Tracker Constants
TR_LOADED="loaded"          # [Boolean] If tracker has been loaded for player
TR_FIRST="first"            # [Boolean] True after first spawn
TR_COUNT="count"            # [String Set] Holds the names of different weapons the player has scored kills with
TR_DELAY="delay"            # [Integer]/[Integer List] The ammount of time before a switch completes.
                            #                          Holds a list when using rolling switches, each element holdinf a time for a weapon.
TR_DELAY_START="start"      # [Integer] When the player started waiting for a switch to complete.
TR_FREE="free"              # [Integer] Tracks a player's award switch. -1 if no switch earned. Otherwise it holds the time the switch is valid for.
TR_KNIFE="knife"            # [Boolean] If the player has the Hunting Knife.
TR_THROW="throw"            # [Boolean] If the player has the Throwing Knife.
TR_KLOBB="klobb"            # [Boolean] If the player has the Klobb.
TR_SPEED="speed"            # [Integer] Holds the player's currnet speed.
TR_SWITCH="switch"          # [Integer] Number of switches the player has previously performed.
TR_AWARD_RESPAWN="award"    # [Boolean] If the player can respawn with an award weapon.
TR_HIDE="hide"              # [Boolean] If the player is hidden from radar.
TR_LIST="list"              # [List Weapon Data] List of weapons the player has spawned with. 
TR_CATS="cats"              # [Set String] Set of Weapon Categories the player will spawn with.
TR_CARRY="carry"            # [Boolean] If the player has opted to carry their currnet weapon set over to their next life.
TR_CAN_CARRY="can-carry"    # [Boolean] If the player is able to carry their current weapon set over to their next life.
TR_CARRY_TIME="ctime"       # [Integer] Time, after which, the player will lose the option to carry for their current life.
TR_KILL_RECORDED="kill"     # [Boolean] True if a player has scored at least 1 kill with their current life.
TR_DEAD_TIME="dtime"        # [Integer[ Time at which the player was last killed.

# <3.2> Command List
CMD_SWAP=["!","!voodoo"]
CMD_MSG=["!gesrocks"]

# <3.3> Message Text
MSG_CLEAR=_("")
MSG_NOTIFY=_("!voodoo to switch weapons.\n!gesrocks to see your weapons.\nSwitching also gives an armor bonus.")
MSG_RADAR=_("You have been hidden from Radar.")
MSG_KNIFE=_("Knives count more for bonuses.\nYou can switch weapons faster.")
MSG_SWAP=_("You have scored a hat trick.\nYou can perform a Weapon Swap at no delay.")
MSG_SWITCH=_("Switching Weapons")
MSG_STANDBY=_("Switching Weapons, please stand by...")
MSG_CARRY=_("You will respawn with the same weapons.")
MSG_UNCARRY=_("You will respawn with different weapons.")
MSG_SWITCH_COMPLETE=_("Switch Complete")
MSG_REMINDER="You can type \"!voodoo\" to respawn with the same weapons you died with."
SMSG_REMAIN_SINGLE="Score a kill with a different weapon for a bonus."
TMSG_REMAIN=Template("Score kills with $n more different weapons for a bonus.")
TMSG_BONUS_KILL=Template("^7$player^1 has scored a hat trick. (+$count kill$pl)")
TMSG_WEAPON_DISPLAY=Template("Slot $n: $slot")
TMSG_ROLLING_SWITCH=Template("[$n/3] Switching Weapons, please stand by...")
TMSG_SPEED=Template("Your Speed has been increased by %%$mod.")
TMSG_DAMAGE=Template("Your Damage has been increased by %%$mod.")

# <3.4> Colors
COLOR_WHITE=GEUtil.Color(255,255,255,255)
COLOR_ALPHA=GEUtil.Color(220,220,220,1)
COLOR_SWITCH=GEUtil.Color(220,220,220,255)
COLOR_NOTIFY=GEUtil.Color(166,94,46,255)

# <3.5> Message Channels
CHANNEL_RADAR=1
CHANNEL_SWITCH=2
CHANNEL_POWER=3
CHANNEL_WEAPON=4
CHANNEL_AWARD=5

# <3.6> Hud Bar Indexes
BAR_INDEX_SWITCH_TITLE=1
BAR_INDEX_SWITCH_BAR=2

# <3.7> Weapon Data
weaponList={
    "ubers":[(1,1),[[(1,1),"weapon_golden_pp7",0],[(2,2),"weapon_golden_gun",0],[(3,5),"weapon_grenade_launcher",3],[(6,8),"weapon_moonraker",0],[(8,10),"weapon_rcp90",0]],10],
    "handguns":[(2,21),[[(1,1),"weapon_silver_pp7",7],[(2,3),"weapon_cmag",10],[(4,6),"weapon_dd44",24],[(7,9),"weapon_pp7",31]],9],
    "shotguns":[(22,37),[[(1,1),"weapon_auto_shotgun",5],[(2,4),"weapon_shotgun",20]],4],
    "autos":[(38,60),[[(1,4),"weapon_phantom",50],[(5,8),"weapon_zmg",32],[(8,11),"weapon_d5k",90]],11],
    "rifles":[(61,82),[[(1,31),"weapon_sniper_rifle",8],[(32,45),"weapon_ar33",30],[(46,80),"weapon_kf7",90]],80],
    "explosives":[(83,94),[[(1,1),"weapon_timedmine",7],[(2,2),"weapon_proximitymine",5],[(3,7),"weapon_remotemine",5],[(8,9),"weapon_grenade",9], [(10,14),"weapon_rocket_launcher",1]],14],
    "specials":[(95,100),[[(1,4),"weapon_klobb",200],[(5,8),"weapon_knife",0],[(9,10),"weapon_knife_throwing",9],[(10,13),"weapon_pp7_silenced",56],[(14,15),"weapon_d5k_silenced",210]],15]}

#Ammo Counts by Entropy
entropyList={
    "ubers":[(1,1),[[(1,1),"weapon_golden_pp7",0],[(2,2),"weapon_golden_gun",1],[(3,5),"weapon_grenade_launcher",3],[(6,8),"weapon_moonraker",0],[(8,10),"weapon_rcp90",0]],10],
    "handguns":[(2,21),[[(1,1),"weapon_silver_pp7",7],[(2,3),"weapon_cmag",7],[(4,6),"weapon_dd44",24],[(7,9),"weapon_pp7",21]],9],
    "shotguns":[(22,37),[[(1,1),"weapon_auto_shotgun",5],[(2,4),"weapon_shotgun",15]],4],
    "autos":[(38,60),[[(1,4),"weapon_phantom",50],[(5,8),"weapon_zmg",32],[(8,11),"weapon_d5k",30]],11],
    "rifles":[(61,82),[[(1,31),"weapon_sniper_rifle",16],[(32,45),"weapon_ar33",30],[(46,80),"weapon_kf7",60]],80],
    "explosives":[(83,94),[[(1,1),"weapon_timedmine",7],[(2,2),"weapon_proximitymine",3],[(3,7),"weapon_remotemine",4],[(8,9),"weapon_grenade",9], [(10,14),"weapon_rocket_launcher",1]],14],
    "specials":[(95,100),[[(1,4),"weapon_klobb",20],[(5,8),"weapon_knife",0],[(9,10),"weapon_knife_throwing",9],[(10,13),"weapon_pp7_silenced",7],[(14,15),"weapon_d5k_silenced",30]],15]}

weapon_names={
    "weapon_moonraker":"Moonraker Laser",
    "weapon_golden_pp7":"Golden PP7",
    "weapon_golden_gun":"Golden Gun",
    "weapon_silver_pp7":"Silver PP7",
    "weapon_rcp90":"RC-P90",
    "weapon_cmag":"Cougar Magnum",
    "weapon_dd44":"DD44 Dostovei",
    "weapon_pp7":"PP7",
    "weapon_auto_shotgun":"Automatic Shotgun",
    "weapon_shotgun":"Shotgun",
    "weapon_rocket_launcher":"Rocket Launcher",
    "weapon_grenade_launcher":"Grenade Launcher",
    "weapon_phantom":"Phantom",
    "weapon_zmg":"ZMG",
    "weapon_d5k":"D5K Deutsche",
    "weapon_sniper_rifle":"Sniper Rifle",
    "weapon_ar33":"AR33",
    "weapon_kf7":"KF7 Soviet",
    "weapon_timedmine":"Timed Mines",
    "weapon_proximitymine":"Proximity Mines",
    "weapon_remotemine":"Remote Mines",
    "weapon_grenade":"Hand Grenades",
    "weapon_klobb":"Klobb",
    "weapon_knife":"Hunting Knife",
    "weapon_knife_throwing":"Throwing Knife",
    "weapon_pp7_silenced":"PP7 (Silenced)",
    "weapon_d5k_silenced":"D5K (Silenced)",
    "weapon_slappers":"Slappers"
}

# Used to set the matching tracker variables when the named weapon is spawned.
special={
    "weapon_knife":TR_KNIFE,
    "weapon_knife_throwing":TR_THROW,
    "weapon_klobb":TR_KLOBB,
    "weapon_pp7_silenced":TR_HIDE,
    "weapon_d5k_silenced":TR_HIDE
}

# <4> FUCNTIONS
# <4.1> Implication
def impl(a,c):
    """
        Returns a->c (a implies c).
        
        :param a: The antecedent
        :param c: The consequent
        :returns: True if a->c

        Truth Table:
        A|C| 
        -----
        F|F|T
        T|F|F
        F|T|T
        T|T|T
    """
    return not a or c


class Cache( GEScenario ):
    # <4.2> CLASS FUNCTIONS

    # <4.2.1> __init__
    def __init__( self ):
        GEScenario.__init__( self )
        self.rstate=random.getstate()
        self.pltracker=GEPlayerTracker(self)
        self.weapons={}
        
        # Load CVar Values
        self.cvars={
            "cs_enableteamplay":0,
            "cs_disableroundtime":0,
            "cs_intromessage":5,
            "cs_disablelaser":1,
            "cs_entropyammo":1,
            "cs_carrytime":15,
            "cs_scorelimit":0,
            "cs_enablenotifyremaining":1,
            "cs_enableslappers":0,
            "cs_enableswitching":1,
            "cs_enableawardswitch":1,
            "cs_enableawardswitchcarry":1,
            "cs_enableeffects":1,
            "cs_speedmultiplier":1.15,
            "cs_damagemultiplier":1.5,
            "cs_enablestealth":1,
            "cs_enableubernormal":1,
            "cs_awardautoswitch":0,
            "cs_awardswitchclass":1,
            "cs_switchdelay":10,
            "cs_switchtimepenalty":1,
            "cs_nodelayonkill":0,
            "cs_knifeswitchdelay":5,
            "cs_awardswitchduration":10,
            "cs_switchdefenseless":1,
            "cs_switchslappers":1,
            "cs_switchonslapkill":1,
            "cs_enablearmorbonus":2,
            "cs_switcharmorbonus":0.25,
            "cs_enablearmoronaward":1,
            "cs_enablebonusarmorstacking":1,
            "cs_bonus":1,
            
            # CVars for Rolling Switch
            "cs_rs_enable":0,
            "cs_rs_1":2,
            "cs_rs_2":3,
            "cs_rs_3":3,
            "cs_rs_k1":0,
            "cs_rs_k2":3,
            "cs_rs_k3":6,
            "cs_rs_x1":0,
            "cs_rs_x2":2,
            "cs_rs_x3":4,
            "cs_rs_p1":1,
            "cs_rs_p2":2,
            "cs_rs_p3":3,
        }
    
    # <4.2.2> GetPrintName
    def GetPrintName( self ):
        return "Cache"

    # <4.2.3> GetScenarioHelp
    def GetScenarioHelp( self, help_obj ):
        help_obj.SetDescription("Every player spawns with 3 different weapons, selected at random with weighted probabilites. These weapons will change on each respawn.\n\nSaying \"!voodoo\" or \"!gesrocks\" in chat will give you a new set of Weapons after a 10 second delay, however you will have to go without any Weapons for the duration of the switch.\n\nIf you score a kill with 3 different Weapons, you will receive an additional point. Additionally, you will be able to switch your Weapons once at no delay, you must do this within 10 seconds of receiving this bonus. If you do so, you will also receive a high quality weapon. If you die before using this switch, you will receive your quality weapon on respawn.")

    # <4.2.4> GetGameDescription
    def GetGameDescription( self ):
        return "Team Cache" if GERules.IsTeamplay() else "Cache"

    # <4.2.5> GetTeamPlay
    def GetTeamPlay( self ):
        return Glb.TEAMPLAY_TOGGLE if self.cvars["cs_enableteamplay"] else Glb.TEAMPLAY_NONE

    # <4.2.6> OnLoadGamePlay
    def OnLoadGamePlay(self):
        self.u_CreateCVars()

    # <4.2.7> OnUnloadGamePlay
    def OnUnloadGamePlay(self):
        super( Cache, self ).OnUnloadGamePlay()

    # <4.2.8> OnCVarChanged
    def OnCVarChanged(self,name,old,new):
      if name in self.cvars:
        try:
          self.cvars[name]=int(new)
        except ValueError:
            try:
                self.cvars[name]=float(new)
            except ValueError:
                print("ERROR: \"" + name + "\" is not a number")

    # <4.2.9> OnRoundBegin
    def OnRoundBegin( self ):
        GEScenario.OnRoundBegin( self )

        GERules.EnableSuperfluousAreas()
        GERules.DisableWeaponSpawns()
        GERules.DisableAmmoSpawns()
        GERules.AllowRoundTimer(not self.cvars["cs_disableroundtime"])

        self.weapons = entropyList if self.cvars["cs_entropyammo"] else weaponList
        print(self.cvars["cs_entropyammo"])
        print(self.weapons==weaponList)

    # <4.2.10> OnPlayerSpawn
    def OnPlayerSpawn( self, player ):
        t=self.pltracker[player]
        t[TR_SWITCH]=0
        self.wp_GiveWeapons(player,TR_AWARD_RESPAWN in t and t[TR_AWARD_RESPAWN])
        if not TR_FIRST in t:
            notify=self.cvars["cs_intromessage"]
            t[TR_FIRST]=True
            if notify:
              GEUtil.HudMessage(player,MSG_NOTIFY,-1,0.50,COLOR_NOTIFY,notify,CHANNEL_AWARD)

    # <4.2.11> OnPlayerKilled
    def OnPlayerKilled( self, victim, killer, weapon ):
        # Let the base scenario behavior handle scoring so we can just worry about the gun swap mechanics.
        GEScenario.OnPlayerKilled( self, victim, killer, weapon )
        self.p_OnKilled(victim)
        if victim==killer or (GERules.IsTeamplay() and killer.GetTeamNumber()==victim.GetTeamNumber()):
            return
        self.p_OnKill(killer,weapon)

    # <4.2.12> OnPlayerSay
    def OnPlayerSay(self,player,text):
        """
            Used to handle requests to switch weapons from the player.
            The command '!' was originally chosen because it should be quick to type, but hard to type by accident.
            !gesrocks or !voodoo can also be used as they can be bound to buttons.
        """
        hmsg=partial(GEUtil.HudMessage,player)
        if text in CMD_SWAP:
            if player.IsDead():
                self.sw_SetCarry(player)
            else:
                self.sw_SwitchWeapons(player)
            return True
        elif text in CMD_MSG:
            if not player.IsDead():
                self.ui_PrintLoadout(player,"Your current weapons:")
            return True
        elif text=="!test messages":
            GEUtil.HudMessage(player,MSG_SWAP,-1,0.71,COLOR_WHITE,self.cvars["cs_awardswitchduration"] or 5.0,CHANNEL_AWARD)
            hmsg(MSG_KNIFE,-1,0.65,COLOR_WHITE,3.0,CHANNEL_POWER)
            hmsg(MSG_RADAR,-1,0.75,COLOR_WHITE,3.0,CHANNEL_RADAR)
            self.ui_PrintLoadout(player)
            return True
        elif text=="!reset first":
          del self.pltracker[player][TR_FIRST]
          return True
        elif text=="!cheat":
            if GEUtil.GetCVarValue("sv_cheats")=="1":
                player.StripAllWeapons()
                self.wp_SetCarryLoadout(player,[("weapon_golden_gun",500),("weapon_golden_pp7",500),("weapon_rcp90",500)])
                player.SetArmor(player.GetMaxArmor())
                player.SetHealth(player.GetMaxHealth())
            return True
        elif text=="!noarmor":
            if GEUtil.GetCVarValue("sv_cheats")=="1":
                player.SetArmor(0)
                return True
        elif text=="!heal":
            if GEUtil.GetCVarValue("sv_cheats")=="1":
                player.SetHealth(player.GetMaxHealth())
                return True
        elif text=="!reset time":
            if GEUtil.GetCVarValue("sv_cheats")=="1":
                self.pltracker[player][TR_SWITCH]=0
                return True
        elif text=="!noweapons":
            if GEUtil.GetCVarValue("sv_cheats")=="1":
                for pl in GetPlayers():
                    pl.StripAllWeapons()
                return True
        elif text=="!award":
            if GEUtil.GetCVarValue("sv_cheats")=="1":
                self.pltracker[player][TR_FREE]=GetTime()+500
                return True
        return False

    # <4.2.13> OnThink
    def OnThink(self):
        for player in GetPlayers():

            delay=self.pltracker[player][TR_DELAY] if TR_DELAY in self.pltracker[player] else -1
            if not delay==-1:
                self.ui_UpdateSwitchBar(player)

            isRolling=isinstance(delay,list)
            timeToCheck=delay[0] if isRolling else delay

            if not self.t_CheckTimer(timeToCheck):
                continue

            self.sw_SwitchArmor(player,isRolling)
            if isRolling:
                self.sw_RollingSwitch(player,delay)
            else:
                self.wp_GiveWeapons(player,False)

    # <4.2.14> CanMatchEnd
    def CanMatchEnd(self):
       return impl(self.cvars["cs_disableroundtime"],GERules.GetNumActivePlayers()<2)

# ---------------------------
# GAMEPLAY SPECIFIC FUNCTIONS
# ---------------------------
    # <4.3> UTILITY FUNCTIONS

    # <4.3.1> u_GetRandom
    def u_GetRandom(self,mx):
        """
            Generates a random number between 1 and a given maximum value.
            randint isn't directly called, because in playtesting the generator kept giving the same result if used by different function calls (or something like that)
            This function fixes that issue by keeping track of the rng's state after each call.

            :param mx:  The largest number that can be generated
            :return:    A random number between 1 and mx
        """
        random.setstate(self.rstate)
        n=random.randint(1,mx)
        self.rstate=random.getstate()
        return n

    # <4.3.2> u_CheckRange
    def u_CheckRange(self,n,mn,mx):
        """
           Checks if a number is between a given minimum and maximum value (inclusive)
            
            :param n:   The number to check
            :param mn:  The lower bound for the number
            :param mx:  The upper bound for the number
            
            :return: True if n in greater or equal to mn, but less or equal to mx. False otherwise.
        """
        return mx>=n>=mn

    # <4.3.3> u_CreateCVars
    def u_CreateCVars(self):
        """
            Is used to initialize all the CVar values for this mode to their default stats at the start of the game.
        """
        f=self.CreateCVar
        f("cs_enableteamplay","1","If 1, allows Team Play during game.")
        f("cs_disableroundtime","0","If 1 at the start of the game, disables the round timer.")
        f("cs_intromessage","5","Time, in seconds, to display the intro message upon first spawn. Set to 0 to disable.")
        f("cs_disablelaser","1","If 1, the Moonraker Laser will not appear.")
        f("cs_entropyammo","1","If 1, uses Entropy-Soldier's suggested ammo count instead of the defaults.")
        f("cs_carrytime","15","Ammount of time a player has until they can not carry their current weapon set.")
        f("cs_scorelimit","0","The number of kills a player needs to win the round (0 removes the limit).")
        f("cs_enablenotifyremaining","1","If 1, player will be notified how many kills they need for a bonus each time they score a kill.")
        f("cs_enableslappers","0","If 1, gives players Slappers in addition to normal weapons.")
        f("cs_enableswitching","1","If 1, allows players to use the message '!', '!voodoo', or '!gesrocks' to switch weapons.")
        f("cs_enableawardswitch","1","If 1, player is given a free switch after scoring a kill with 3 different weapons.")
        f("cs_awardswitchclass","1","If 1, an award switch will ensure the player is given a weapon from the 'ubers' list.")
        f("cs_enableawardswitchcarry","1","If 1, player will recieve a weapon from the 'ubers' list if they die before using their award switch")
        f("cs_enableeffects","1","If 1, Throwing Knives; Hunting Knives; Klobbs; and Silenced Weapons will provide unique effects.")
        f("cs_enablestealth","1","If 1, Silenced Weapons will hide you from radar.")
        f("cs_speedmultiplier","1.15","Player's speed multiplier when they spawn with the Hunting Knife.")
        f("cs_damagemultiplier","1.5","Player's damage multiplier when they spawn with the Klobb.")
        f("cs_enableubernormal","1","If 1, allows players to get weapons from the 'ubers' list without needing an award switch.")
        f("cs_awardautoswitch","0","If 1, player will automatically use their award switch when the recieve it.")
        f("cs_switchdelay","10","Time, in seconds, between requesting a weapon switch and recieving new weapons. (Doesn't apply to award switches.")
        f("cs_switchtimepenalty","1","Time, in seconds, to add on to the switch delay after each switch. Numbers from 0 to 9 are valid. Higher numbers means that each digit (from left to right) will be added after the matching switch (So 12 will add 1 second on the 2nd switch attempt, and 2 seconds to the 3rd and later). These are cumulative. To stop accumulation, use a 0.")
        f("cs_nodelayonkill","0","If the switch delay should be ignored if the play has scored a kill.")
        f("cs_knifeswitchdelay","3","Time, in seconds, between requesting a weapon switch  and recieving new weapons. (Used instead of the normal delay if the player spawns with Throwing Knives).")
        f("cs_awardswitchduration","10","Time, in seconds, after making a kill with 3 different weapons, a player has to claim their award switch.")
        f("cs_switchdefenseless","1","If 1, player will lose all weapons after requesting a weapon switch, untill the switch is complete.")
        f("cs_switchslappers","1","If 1, player will gain slappers when requesting a weapon switch.")
        f("cs_switchonslapkill","1","If 1, scoring a slappers kill while you're performing a weapon switch will immediately cancel the waiting period.")
        f("cs_enablearmorbonus","2","If 1, armor will be awarded when switching after making kills. If 2, armor will be awarded after the 2nd and 3rd weapon is given during a switch.")
        f("cs_switcharmorbonus","0.25","Armor increase, as a multiple of the max armor a player can have, that will be added per unique weapon kill when a player switches.")
        f("cs_enablearmoronaward","1","If 1, then armor is also awarded with an award switch.")
        f("cs_enablebonusarmorstacking","1","If 1, then armor gained through bonuses can stack on to the player's armor total. If 0, then the player's armor will be set to their bonus, or their current armor, whichever is higher")
        f("cs_bonus","1","Number of additional points to award after scoring a kill with 3 different weapons.")

        #CVars for Rolling Switch
        f("cs_rs_enable","1","If 1, enables rolling switch option.")
        f("cs_rs_1","2","Time, in seconds, for the first weapon to be given during a Rolling Switch.")
        f("cs_rs_2","5","Time, in seconds, for the second weapon to be given during a Rolling Switch.")
        f("cs_rs_3","8","Time, in seconds, for the third weapon to be given during a Rolling Switch.")
        f("cs_rs_k1","0","Time, in seconds, for the first weapon to be given during a Rolling Switch, if you started with Throwing Knives.")
        f("cs_rs_k2","2","Time, in seconds, for the second weapon to be given during a Rolling Switch, if you started with Throwing Knives.")
        f("cs_rs_k3","4","Time, in seconds, for the third weapon to be given during a Rolling Switch, if you started with Throwing Knives.")
        f("cs_rs_x1","0","Time, in seconds, for the first weapon to be given during a Rolling Switch if you have made a kill.")
        f("cs_rs_x2","2","Time, in seconds, for the second weapon to be given during a Rolling Switch if you have made a kill.")
        f("cs_rs_x3","4","Time, in seconds, for the third weapon to be given during a Rolling Switch if you have made a kill.")
        f("cs_rs_p1","1","Time, in seconds, to add on to the first weapon's rolling switch delay after each switch. Numbers from 0 to 9 are valid. Higher numbers means that each digit (from left to right) will be added after the matching switch (So 12 will add 1 second on the 2nd switch attempt, and 2 seconds to the 3rd and later). These are cumulative. To stop accumulation, use a 0.")
        f("cs_rs_p2","2","Time, in seconds, to add on to the second weapon's rolling switch delay after each switch. Numbers from 0 to 9 are valid. Higher numbers means that each digit (from left to right) will be added after the matching switch (So 12 will add 1 second on the 2nd switch attempt, and 2 seconds to the 3rd and later). These are cumulative. To stop accumulation, use a 0.")
        f("cs_rs_p3","3","Time, in seconds, to add on to the third weapon's rolling switch delay after each switch. Numbers from 0 to 9 are valid. Higher numbers means that each digit (from left to right) will be added after the matching switch (So 12 will add 1 second on the 2nd switch attempt, and 2 seconds to the 3rd and later). These are cumulative. To stop accumulation, use a 0.")

    # <4.3.4> u_CheckKillEnd
    def u_CheckKillEnd(self,player):
        """
            Ends the round if the given player has enough scored kills

            :param player: The player that will be the target of this function
        """
        lim=self.cvars["cs_scorelimit"]
        ct=player.GetRoundScore()
        if lim and ct>=lim:
            GERules.EndRound()

    # <4.3.5> u_GetDigit
    def u_GetDigit(self,number,digit):
        """
            Prints the given digit of the given number

            :param number:  The number to get a digit from.
            :param digit:   The digit (starting from the right) to get from the given number.

            :return:        THe given digit from the given number.
        """
        if digit==0:
            return 0
        num=str(number)
        if (digit>len(num)):
            digit=len(num)
        elif digit<=0:
            digit=1
        digit-=1
        return int(num[digit])

    # <4.4> TIMER FUCNTIONS
    # <4.4.1> t_OffsetTime
    def t_OffsetTime(self,offset):
        """
            Adds an offset to the current server time.

            :param offset:  The ammount of time to add/subtract
            :return:        What the time will be after the offset is applied
        """
        return GetTime()+offset

    # <4.4.2> t_CheckTimer
    def t_CheckTimer(self,time,isLater=True,checkTime=None):
        """
            Checks to see if a given time has passed, and that the time exists (is not a negative number).
            
            :param time:        The time to be compared to the current game time
            :param isLater:     If True, checks if the time has passed. If False checks if the time has not passed.
            :param checkTime:   If provided, this time is used instead of the current game time
            :return:            True if the game time is at or ahead of the given time, and false otherwise.
        """
        if checkTime is None:
            checkTime=GetTime()
        if isLater:
            return checkTime>=time>=0
        else:
            return 0<=time>=checkTime

    # <4.5> PLAYER FUNCTIONS
    # 84.5.1> p_ResetPlayer
    def p_ResetPlayer(self,player):
        """
            Resets the player tracker values, along with speed and damage multipliers for the given player.

            :param player: The player to reset.
        """
        t=self.pltracker[player]
        t[TR_LOADED]=True
        t[TR_COUNT]=set()
        t[TR_DELAY]=-1
        t[TR_DELAY_START]=-1
        t[TR_FREE]=-1
        t[TR_KNIFE]=False
        t[TR_THROW]=False
        t[TR_HIDE]=False
        t[TR_KLOBB]=False
        t[TR_SPEED]=1
        t[TR_AWARD_RESPAWN]=False
        t[TR_LIST]=[]
        t[TR_CATS]=set()
        t[TR_CARRY]=False
        t[TR_CAN_CARRY]=False
        t[TR_CARRY_TIME]=self.t_OffsetTime(self.cvars["cs_carrytime"]) if self.cvars["cs_carrytime"] else -1
        t[TR_KILL_RECORDED]=False
        player.SetDamageMultiplier(1)
        player.SetSpeedMultiplier(1)
        # Clears out messages
        for n in range(1,6):
            GEUtil.HudMessage(player,MSG_CLEAR,-1,0.25,COLOR_WHITE,0,n)
        self.ui_DestroySwitchBar(player)

    # <4.5.2> p_OnKilled
    def p_OnKilled(self,player):
        """
            Used to handle players that die. Removing their weapons, and setting tracker variables.
            
            :param player: The player that died
        """
        for n in range(1,6):
            GEUtil.HudMessage(player,MSG_CLEAR,-1,0,COLOR_WHITE,0,n)
        player.StripAllWeapons()
        tr=self.pltracker[player]
        tr[TR_DELAY]=-1
        tr[TR_DEAD_TIME]=GetTime()
        if self.sw_CheckCarry(player):
            GEUtil.HudMessage(player,MSG_REMINDER,-1,0.25,COLOR_SWITCH,3,CHANNEL_SWITCH)
            tr[TR_CAN_CARRY]=True
        if self.t_CheckTimer(tr[TR_FREE]):
            tr[TR_AWARD_RESPAWN]=self.cvars["cs_enableawardswitchcarry"]

    # <4.5.3> p_OnKill
    def p_OnKill(self,player,weapon):
        """
            Handles the various things that need to happen when a player scores a kill.
            Such as the awarding of additional points, and tracking progress towards bonuses.
            
            :param player: The player that made the kill
            :param weapon: The weapon used for the kill
        """
        if not (player and weapon):
            return
        tr=self.pltracker[player]
        prevc=len(tr[TR_COUNT])
        name=weapon.GetClassname().lower()
        if name=="weapon_knife_throwing" and self.cvars["cs_enableeffects"]:
            tr[TR_COUNT].add("bonus")
        tr[TR_COUNT].add(weapon.GetClassname())
        tr[TR_KILL_RECORDED]=True
        self.ui_NotifyKillsRemaining(player,prevc,name)
        self.sw_AwardBonus(player)
        self.u_CheckKillEnd(player)
        self.sw_SlaperKill(player,weapon)

    # <4.6> WEAPON FUNCTIONS

    # <4.6.1> wp_GiveWeapons
    def wp_GiveWeapons(self,player,award,count=3):
        """
            Is used to set the player's Weapons on respawn and if they manually switch.
            The given player will be assigned 3 random weapons from different categories.
            Special benefits of certain weapons is also handled.

            :param player:  The player that will be the target of this function.
            :param award:   If True, the player will be assured a weapon from the 'ubers' list.
        """
        if TR_LOADED not in self.pltracker[player]:
            self.p_ResetPlayer(player)
        carry=self.pltracker[player][TR_CARRY]
        weapons=self.pltracker[player][TR_LIST]
        player.StripAllWeapons()
        if self.cvars["cs_enableslappers"]:
            player.GiveNamedWeapon("weapon_slappers",0)
        self.p_ResetPlayer(player)
        if carry:
            self.wp_SetCarryLoadout(player,weapons)
        else:
            self.wp_AssignWeapons(player,award,count)
        self.ui_PrintLoadout(player)
        if not self.cvars["cs_enableeffects"]:
            return
        self.wp_SpecialWeapons(player)
        self.wp_StealthWeapons(player)
    
    # <4.6.2> wp_AssignWeapons
    def wp_AssignWeapons(self,player,award,count):
        """
            Used to handle just the assignment of a wp_GiveWeapons call.
            Weapons are assigned, but any weapons with special traits wont have those applied here.
            See wp_GiveWeapons for more info.

            :param player:  The player that will be the target of the function.
            :param award:   If True, the player will be assured a weapon from the 'ubers' list.
            :param count:   The number of weapons to be assigned. (NOTE: You really shouldn't use anything other than 3)
        """
        cats=[]
        for n in range(count):
            cats.append(self.wp_AssignSingleWeapon(player,cats,award and self.cvars["cs_awardswitchclass"])[0])
            award=False
    
    # <4.6.3> wp_AssignSingleWeapon
    def wp_AssignSingleWeapon(self,player,used_cats,award=False):
        """
            Used to assign a single weapon to the player.
            Weapons are assigned, but any weapons with special tratis wont have those applied here.
            
            :param player:      The player that will be the target of the function.
            :param award:       Set to True to ensure an Ubers Weapon.
            :param used_cats:   A list of weapons categories that the player has already used.

            :return:            The category of the selected weapon.
        """
        cat=None
        while (not cat) or (cat in used_cats):
            cat=self.wp_GenWeaponCat(award)
        w,a=self.wp_GenWeapon(cat)
        self.pltracker[player][TR_LIST].append((w,a))
        player.GiveNamedWeapon(w,a)
        if w in special:
            self.pltracker[player][special[w]]=True
        return cat,w

    # <4.6.4> wp_GenWeaponCat
    def wp_GenWeaponCat(self,award):
        """
            Generates a random weapon category from the Weapon List.
            choice(self.weapons) cannot be used because of the weights assigned to each category.
            
            :return: Randomly selected category form self.weapons.
        """
        if award:
            return "ubers"
        n=self.u_GetRandom(100)
        cat=next(k for k in self.weapons if self.u_CheckRange(n,*self.weapons[k][0]))
        if not impl(cat=="ubers",self.cvars["cs_enableubernormal"]):
            return self.wp_GenWeaponCat(False)
        return cat

    # <4.6.5> wp_GenWeapon
    def wp_GenWeapon(self,cat):
        """
            Generates a random weapon of the given category from self.weapons.
            As with wp_GenWeaponCat, the choice function cannot be used because the weapons are weighted.
            
            :param cat: The category to generate from.
            :return:    A tuple, the first value contains the weapon, and the second value contains the ammount of ammo to assign to that weapon.
                        (weaponName,ammo)
        """
        d=self.weapons[cat][1]
        mx=self.weapons[cat][2]
        n=self.u_GetRandom(mx)
        w=next((v[1],v[2]) for v in d if self.u_CheckRange(n,*v[0]))
        if w[0]=="weapon_moonraker" and self.cvars["cs_disablelaser"]:
            print("Reroll")
            return self.wp_GenWeapon(cat)
        else:
            return w

    # <4.6.6> wp_StealthWeapons
    def wp_StealthWeapons(self,player):
        """
            Is used to check if a player spawns with a stealth weapon, and if so, hides that player from Radar.
            
            :param player: The player to check
        """
        r=GERules.GetRadar()
        if self.pltracker[player][TR_HIDE] and self.cvars["cs_enablestealth"]:
            #Credit to Entropy-Soldier for this trick
            r.AddRadarContact(player,Glb.RADAR_TYPE_PLAYER,False,"",COLOR_ALPHA)
            GEUtil.HudMessage(player,MSG_RADAR,-1,0.75,COLOR_WHITE,3.0,CHANNEL_RADAR)
        else:
            r.DropRadarContact(player)
    
    # <4.6.7> wp_SpecialWeapons
    def wp_SpecialWeapons(self,player):
        """
            Is used to see if a player spawned with a weapon with a special bonus.
            If so, this also sets the bonus on the player, and sends them a notification.
            
            :param player:  The player that will be the target of this function
        """
        hmsg=lambda m:GEUtil.HudMessage(player,m,-1,0.65,COLOR_WHITE,3.0,CHANNEL_POWER)
        if self.pltracker[player][TR_KNIFE]:
            mod=self.cvars["cs_speedmultiplier"]
            player.SetSpeedMultiplier(mod)
            self.pltracker[player][TR_SPEED]=mod
            hmsg(_(TMSG_SPEED.substitute(mod=round((mod-1)*100))))
        elif self.pltracker[player][TR_KLOBB]:
            mod=self.cvars["cs_damagemultiplier"]
            player.SetDamageMultiplier(mod)
            hmsg(_(TMSG_DAMAGE.substitute(mod=round((mod-1)*100))))
        elif self.pltracker[player][TR_THROW]:
            hmsg(MSG_KNIFE)

    # <4.6.8> wp_SetCarryLoadout
    def wp_SetCarryLoadout(self,player,weapons):
        """
            Sets a players loadout to the provided loadout.
            Used to return a player's weapons if they carry them over.

            :param player:  The player that will be the target of this function
            :param weapons: A list of weapons and their ammo count to equip a player with.
                            For example [("weapon_pp7",31),("weapon_kf7",90),("weapon_shotgun",15)]
        """
        self.pltracker[player][TR_LIST]=weapons
        for weapon,ammo in weapons:
            player.GiveNamedWeapon(weapon,ammo)
            if weapon in special:
              self.pltracker[player][special[weapon]]=True

    # <4.7> WEAPON SWITCHING FUCNTIONS

    # <4.7.1> sw_AwardBonus
    def sw_AwardBonus(self,player):
        """
            Is used to check for, and give a player their award for scoring a kill with each of their Weapons (if any has been set in the CVar values).

            :param player:  The player that will be the target of the function.
        """
        n=3-len(self.pltracker[player][TR_COUNT])
        if n<=0:
            if self.cvars["cs_bonus"]:
                player.AddRoundScore(self.cvars["cs_bonus"])
                single=self.cvars["cs_bonus"]==1
                count=self.cvars["cs_bonus"]
                msg=_(TMSG_BONUS_KILL.substitute(player=player.GetCleanPlayerName(),count=count,pl="" if single else "s"))
                GEUtil.PostDeathMessage(msg)
            self.sw_GiveAwardSwitch(player)
            self.pltracker[player][TR_COUNT]=set()
   
    # <4.7.2> sw_GiveAwardSwitch
    def sw_GiveAwardSwitch(self,player):
            """
                Is used to give a player their award switch and handle all the CVars used for this proess.
                
                :param player: The player that requested the switch.
            """
            if self.cvars["cs_enableawardswitch"]:
                t=self.pltracker[player][TR_FREE]
                GEUtil.HudMessage(player,MSG_SWAP,-1,0.71,COLOR_SWITCH,self.cvars["cs_awardswitchduration"] or 5.0,CHANNEL_AWARD)
                self.pltracker[player][TR_FREE]=self.cvars["cs_awardswitchduration"] and self.t_OffsetTime(self.cvars["cs_awardswitchduration"])
                if self.cvars["cs_awardautoswitch"]:
                    self.wp_GiveWeapons(player,True)

    # <4.7.3> sw_SwitchWeapons
    def sw_SwitchWeapons(self,player):
            """
                Switches the players weapons when they request it.
                Handles both a normal switch, and an award switch.

                :param player: The player performing the switch
            """
            if player.IsDead():
                return
            t=self.pltracker[player][TR_FREE]
            delay=self.pltracker[player][TR_DELAY]
            if self.cvars["cs_enableawardswitch"] and impl(t,self.t_CheckTimer(t,False)):
                self.sw_SwitchArmor(player,False)
                self.pltracker[player][TR_SWITCH]=0
                self.wp_GiveWeapons(player,True)
            elif self.cvars["cs_enableswitching"] and (isinstance(delay,list) or delay<0):
                if self.cvars["cs_enablearmorbonus"]==1:
                    self.sw_SwitchArmor(player,False)
                self.sw_PerformSwitch(player)

    # <4.7.4> sw_PerformSwitch
    def sw_PerformSwitch(self,player):
            """
                Performs a non-Award Switch of the players weapons.
                This process (normally) involves taking the players weapons for a set time, then after that time, assigning them new weapons.

                :param player: The player performing the switch
            """
            self.pltracker[player][TR_SWITCH]+=1
            rs=not self.cvars["cs_rs_enable"]
            delay=self.sw_GetSwitchDelay(player) if rs else self.sw_GetRollingSwitchDelay(player)
            d=delay if rs else delay[0]
            if self.cvars["cs_switchdefenseless"] or not rs:
                player.StripAllWeapons()
                self.p_ResetPlayer(player)
            if self.cvars["cs_switchslappers"]:
                player.GiveNamedWeapon("weapon_slappers",0)

            self.pltracker[player][TR_DELAY]=delay
            self.pltracker[player][TR_DELAY_START]=GetTime()
            self.ui_CreateSwitchBar(player)

    # <4.7.5> sw_RollingSwitch
    def sw_RollingSwitch(self,player,delay):
        """
            Performs a single step of a Rolling Switch, giving them another of their three weapons.

            :param player:  The player that will be the target of this function
            :param delay:   The delay between gaining weapons
        """
        delay=delay[1:] or -1
        self.pltracker[player][TR_DELAY]=delay
        self.pltracker[player][TR_DELAY_START]=GetTime()
        cats=self.pltracker[player][TR_CATS]
        if not cats:
            player.StripAllWeapons()
        c,w=self.wp_AssignSingleWeapon(player,cats)
        cats.add(c)
        self.ui_PrintLoadout(player)
        self.wp_SpecialWeapons(player)
        self.wp_StealthWeapons(player)
        if isinstance(delay,list):
            self.ui_UpdateSwitchBar(player)
        else:
            self.ui_DestroySwitchBar(player)

    # <4.7.6> sw_GetSwitchDelay
    def sw_GetSwitchDelay(self,player):
        """
            Returns the length of time, in seconds, a weapon switch should take for the player.

            :param player: The player that will be performing the switch
        """
        t=self.pltracker[player]
        c=self.cvars

        isNoKillDelay = t[TR_KILL_RECORDED] and c["cs_nodelayonkill"]
        isKnifeDelay = t[TR_THROW] and c["cs_enableeffects"]

        penalty = sum([self.u_GetDigit(c["cs_switchtimepenalty"],delay) for delay in range(0,t[TR_SWITCH])])

        d=0 if isNoKillDelay else c["cs_knifeswitchdelay"] if isKnifeDelay else c["cs_switchdelay"]
        delay=d+(0 if isNoKillDelay or isKnifeDelay else penalty)

        return self.t_OffsetTime(delay)

    # <4.7.7> sw_GetRollingSwitchDelay
    def sw_GetRollingSwitchDelay(self,player):
        """
            Gets the delay for a rolling switch 
            
            :param player: The player that will be the target of this function.
            :return: A list containing the 3 delay times to be used
        """
        t=self.pltracker[player]
        c=self.cvars

        isNoKillDelay = t[TR_KILL_RECORDED] and c["cs_nodelayonkill"]
        isKnifeDelay = t[TR_THROW] and c["cs_enableeffects"]

        # Which of the cs_rs values to use.
        header="x" if isNoKillDelay else "k" if isKnifeDelay else ""

        # Gets the current time penalty for the given rolling switch slot.
        getPenalty=lambda n:sum([self.u_GetDigit(c["cs_rs_p"+str(n)],digit) for digit in range(0,t[TR_SWITCH])])

        # Generates lists of the switch times for the given header, and time penalties given the current switch count.
        times=[c["cs_rs_"+header+str(n)] for n in range(1,4)]
        penalties=[getPenalty(n) for n in range(1,4)] if header=="" else [0,0,0]

        # Converts the tiems and penalties into the actual delays to be used.
        delays=[self.t_OffsetTime(time+penalty) for time,penalty in zip(times,penalties)]

        return delays

    # <4.7.8> sw_SlaperKill
    def sw_SlaperKill(self,player,weapon):
        """
            Checks if a kill was made with slappers, if that happened while the player was waiting for a weapon swap, and if so, performs that swap now rather than later
            
            :param player:  The player that will be the target of this function
            :param weapon:  The weapon used to make the kill
        """
        rs=not self.cvars["cs_rs_enable"]
        if weapon and weapon.GetClassname()=="weapon_slappers" and self.cvars["cs_switchonslapkill"]:
            tr=self.pltracker[player]
            if rs:
                tr[TR_DELAY]=GetTime() if tr[TR_DELAY]>=0 else -1
            else:
                tr[TR_DELAY]=[GetTime(),GetTime(),GetTime()] if isinstance(tr[TR_DELAY],list) else -1
            # This clears out the switching message
            GEUtil.HudMessage(player,MSG_CLEAR,-1,0.25,COLOR_SWITCH,0,CHANNEL_SWITCH)

    # <4.7.9> sw_SetCarry
    def sw_SetCarry(self,player):
        """
            Used to set a player's weapons to carry over after they request it, if they are able to.
            
            :param player: The player that will be the target of this function
        """
        
        tr=self.pltracker[player]
        if tr[TR_CAN_CARRY]:
            tr[TR_CARRY]^=True
            GEUtil.HudMessage(player,MSG_CARRY if tr[TR_CARRY] else MSG_UNCARRY,-1,0.25,COLOR_SWITCH,3,CHANNEL_SWITCH)

    # <4.7.10> sw_CheckCarry
    def sw_CheckCarry(self,player):
        """
            Used to check if a player is able to carry their current weapon set over.
            To carry weapons, the following conditions must be met

            1. Carry must be enabled (TR_COUNT must not be 0)
            2. The Carry must have been requested within the given time frame (set by cs_carrytime, normally 15)
            3. The player must not have made any kills with their current set.

            :param player: The player to check
            :return: True if all the above conditions are met, and False otherwise.
        """
        t=self.pltracker[player]
        return len(t[TR_COUNT])<=0 and self.t_CheckTimer(t[TR_CARRY_TIME],False,t[TR_DEAD_TIME]) and not t[TR_KILL_RECORDED] and len(t[TR_LIST])>2

    # <4.7.11> sw_SwitchArmor
    def sw_SwitchArmor(self,player,rolling):
        """
            Used to award armor on switching after scoring kills.

            Ammount awarded is based on the cvar cs_switcharmorbonus which is multiplied by the max armor a player can have.
            If cs_enablearmoronaward is set to 1, then armor will be awarded when award switching.
            If cs_enablebonusarmorstacking then this armor will be added to the player's total. Otherwise the player's armor will be set to bonus ammount unless they already have more armor.
        """
        if not self.cvars["cs_enablearmorbonus"]:
            return

        t=self.pltracker[player]
        base=math.floor(player.GetMaxArmor()*self.cvars["cs_switcharmorbonus"])
        bonus=0

        mode=self.cvars["cs_enablearmorbonus"]
        if mode==1:
            bonus=base*len(t[TR_COUNT])
        elif mode==2:
            if rolling:
                bonus=base if len(t[TR_CATS])>0 else 0
            else:
                bonus=base*2

        if impl(t[TR_FREE],self.t_CheckTimer(t[TR_FREE],False)) and self.cvars["cs_enablearmoronaward"]:
            if mode==1:
              bonus=base*3
        if self.cvars["cs_enablebonusarmorstacking"] or mode==2:
            player.SetArmor(player.GetArmor()+bonus)
        else:
            player.SetArmor(max(player.GetArmor(),bonus))

    # <4.8> UI FUNCTIONS
    
    # <4.8.1> ui_CreateSwitchBar
    def ui_CreateSwitchBar(self,player):
        """
            Creates a hud bar that shows the time until weapon switch.

            :param player: The player to show the bar for.
        """
        data=self.ui_GetSwitchBarContent(player)
        #if data[1]==-1:
        #    return
        c=self.cvars
        GEUtil.InitHudProgressBar(player,BAR_INDEX_SWITCH_TITLE,data[0],Glb.HUDPB_TITLEONLY,100,-1,0.24,0,0,COLOR_WHITE,data[1])
        GEUtil.InitHudProgressBar(player,BAR_INDEX_SWITCH_BAR,"",Glb.HUDPB_SHOWBAR,100,0.34,0.27,200,10,COLOR_WHITE,data[1])

    # <4.8.2> ui_UpdateSwitchBar
    def ui_UpdateSwitchBar(self,player):
        """
            Updates a player's hud bar with the current progress towards switching.

            :param player: The player's whose bar should be updated.
        """
        data=self.ui_GetSwitchBarContent(player)
        #if data[1]==-1:
        #    self.ui_DestroySwitchBar(player)
        #    return
        GEUtil.UpdateHudProgressBar(player,BAR_INDEX_SWITCH_TITLE,data[1],data[0])
        GEUtil.UpdateHudProgressBar(player,BAR_INDEX_SWITCH_BAR,data[1])

    # <4.8.3> ui_DestroySwitchBar
    def ui_DestroySwitchBar(self,player):
        """
            Removes the switching progress bar from the given player.

          :param player: The player to remove the bar from.
        """
        GEUtil.RemoveHudProgressBar(player,BAR_INDEX_SWITCH_TITLE)
        GEUtil.RemoveHudProgressBar(player,BAR_INDEX_SWITCH_BAR)

    # <4.8.4> ui_GetSwitchBarContent
    def ui_GetSwitchBarContent(self,player):
        """
            Gets data for a given player's progress bar.

            :param player:  The player to get data for.
            :return:        The title for the player's progress bar, and the percentage the bar should be set to.
        """
        d = self.pltracker[player][TR_DELAY]

        start = self.pltracker[player][TR_DELAY_START]
        time = GetTime()
        delay = d[0] if isinstance(d,list) else d
        remaining = max(delay-time,0)

        progress = time-start
        duration = delay-start
        value = math.floor(progress/duration*100) if duration!=0 else 100
        return "Switch in Progress. {:.1f}s remaining".format(remaining),value

    # <4.8.5> ui_PrintLoadout
    def ui_PrintLoadout(self,player,header=None):
        """
            Prints a given player's current weapon set to their screen.

            :param player: The player that will be the target of this function
            :param header: Message to show before loadout
        """
        ld=[weapon_names[w] for w,a in self.pltracker[player][TR_LIST]]
        s=TMSG_WEAPON_DISPLAY.substitute
        msg=[] if not header else [header]
        for n in range(len(ld)):
            msg.append(s(n=n+1,slot=ld[n]))
        GEUtil.HudMessage(player,_('\n'.join(msg)),-1,0.30,COLOR_WHITE,5,CHANNEL_WEAPON)

    # <4.8.6> ui_NotifyKillsRemaining
    def ui_NotifyKillsRemaining(self,player,prevc,weapon):
        """
            Tells player how many distinct kills they need to make until they earn a bonus.

            :param player:  The player to notify
            :param prevc:   The number of kills they player already had
                                Used so ensure the message only appears when distinct kills are made
            :param weapon:  The weapon used for the kill. Ensures that Slappers don't trigger a message.
        """
        n=3-len(self.pltracker[player][TR_COUNT])
        p=3-prevc
        canNotify=self.cvars["cs_enablenotifyremaining"]
        isKillUnique=n!=p
        notSwitch=n>0
        bonusesEnabled=self.cvars["cs_bonus"]
        awardsEnabled=self.cvars["cs_enableawardswitch"]
        isCountedWeapon=weapon!="weapon_slappers"
        if canNotify and isKillUnique and notSwitch and isCountedWeapon and (bonusesEnabled or awardsEnabled):
            msg=_(TMSG_REMAIN.substitute(n=n) if n>1 else SMSG_REMAIN_SINGLE)
            GEUtil.HudMessage(player,msg,-1,0.71,COLOR_WHITE,3.0,CHANNEL_AWARD)

